@@ -16,6 +16,7 @@
 module Internals (
     module Cont,
     module Posix,
+    module Rule.Pos,
     module Data.Dynamic,
     module Data.Unique,
     module System.Environment,
@@ -40,9 +41,6 @@ module Internals (
     module Data.FiniteMap,
     module Data.IORef,
     module Debug.Trace,
-    module Text.ParserCombinators.Parsec,
-    module Text.ParserCombinators.Parsec.Error,
-    module Text.ParserCombinators.Parsec.Language,
 ) where
 
 import Cont
@@ -73,9 +71,7 @@ import Data.FiniteMap
 import Data.Tree
 import Data.IORef
 import Debug.Trace
-import Text.ParserCombinators.Parsec hiding (parse)
-import Text.ParserCombinators.Parsec.Error hiding (ParseError, errorPos)
-import Text.ParserCombinators.Parsec.Language
+import Rule.Pos
 
 -- Instances.
 instance Show Unique where
