@@ -25,6 +25,7 @@ op0 "!"  = return . VJunc . Junc JNone emptySet . mkSet
 op0 "&"  = return . opJuncAll
 op0 "^"  = return . opJuncOne
 op0 "|"  = return . opJuncAny
+op0 "undef" = \_ -> return VUndef
 op0 "time"  = \_ -> do
     clkt <- liftIO getClockTime
     return $ VInt $ toInteger $ tdSec $ diffClockTimes clkt epochClkT
@@ -35,6 +36,9 @@ op0 s    = \x -> return $ VError ("unimplemented listOp: " ++ s) (Val $ VList x)
 
 op1 :: Ident -> Val -> Eval Val
 op1 "!"    = return . fmapVal not
+op1 "undef" = \mv -> do
+    liftIO $ writeIORef (vCast mv) $ VUndef
+    return VUndef
 op1 "+"    = return . op1Numeric id
 op1 "post:++" = \mv -> do
     val <- readMVal mv
@@ -56,6 +60,7 @@ op1 "--"   = \mv -> do
 op1 "-"    = return . op1Numeric negate
 op1 "~"    = return . VStr . vCast
 op1 "?"    = return . VBool . vCast
+op1 "int"  = return . VInt . vCast
 op1 "*"    = return . VList . vCast
 op1 "**"   = return . VList . map (id $!) . vCast
 op1 "+^"   = return . VInt . (toInteger . (complement :: Word -> Word)) . vCast
@@ -69,7 +74,33 @@ op1 "all"  = return . opJuncAll . vCast
 op1 "one"  = return . opJuncOne . vCast
 op1 "none" = return . VJunc . Junc JNone emptySet . mkSet . vCast
 op1 "perl" = return . VStr . (pretty :: Val -> VStr)
+op1 "require" = \v -> do
+    fileVal <- readMVal v
+    glob <- askGlobal
+    -- XXX - assuming a flat array
+    let Just Symbol{ symExp = Val incAV } = find ((== "@*INC") . symName) glob
+    incVals <- readMVal incAV
+    let incs = map vCast $ vCast incVals
+        file = vCast fileVal
+    requireInc incs file (errMsg file incs)
+    where
+    errMsg file incs = "Can't locate " ++ file ++ " in @INC (@INC contains: " ++ unwords incs ++ ")."
+    requireInc [] _ msg = do
+        return $ VError msg (Val VUndef)
+    requireInc (p:ps) file msg = do
+        let pathName = p ++ "/" ++ file
+        ok <- liftIO $ doesFileExist pathName
+        if (not ok)
+            then requireInc ps file msg
+            else do
+                str <- liftIO $ readFile pathName
+                op1 "eval" $ VStr str
 op1 "eval" = opEval . vCast
+op1 "defined" = \v -> do
+    v <- readMVal v
+    return . VBool $ case v of
+        VUndef  -> False
+        _       -> True    
 op1 "last" = \v -> do
     shiftT $ \_ -> return VUndef
 op1 "return" = \v -> return (VError "cannot return outside a subroutine" (Val v))
@@ -79,15 +110,8 @@ op1 "rand"  = \v -> do
     let x = vCast v
     rand <- liftIO $ randomRIO (0, if x == 0 then 1 else x)
     return $ VNum rand
-op1 "print" = \v -> do
-    v <- readMVal v
-    vals <- mapM readMVal (vCast v)
-    liftIO . putStr . concatMap vCast $ vals
-    return $ VBool True
-op1 "say" = \v -> do
-    op1 "print" v
-    liftIO $ putStrLn ""
-    return $ VBool True
+op1 "print" = op1Print hPutStr
+op1 "say" = op1Print hPutStrLn
 op1 "join"= \v -> do
     v <- readMVal v
     vals <- mapM readMVal (vCast v)
@@ -109,6 +133,11 @@ op1 "sleep" = boolIO sleep
 op1 "mkdir" = boolIO createDirectory
 op1 "rmdir" = boolIO removeDirectory
 op1 "chdir" = boolIO setCurrentDirectory
+op1 "unlink" = \v -> do
+    v <- readMVal v
+    vals <- mapM readMVal (vCast v)
+    rets <- liftIO $ sequence $ map ( iboolIO removeFile ) $ map vCast vals
+    return $ VInt $ sum $ map bool2n rets
 op1 "open" = \v -> do
     fh <- liftIO $ openFile (vCast v) ReadMode
     return $ VHandle fh
@@ -143,15 +172,42 @@ op1 "<>" = \v -> do
         case find ((== "$*STDIN") . symName) glob of
             Just sym | (Val v) <- symExp sym -> return $ vCast v
             _                                -> error "impossible"
-
-op1 ""     = return . (\x -> VError ("unimplemented unaryOp: " ++ "") (Val x))
-
+op1 "ref"  = return . VStr . valType
+op1 "pop"  = op1Pop (last, init)
+op1 "shift"= op1Pop (head, tail)
 op1 s      = return . (\x -> VError ("unimplemented unaryOp: " ++ s) (Val x))
 
-boolIO f v = do
+op1Pop (fPick, fRem) list = do
+    let array = vCast list
+    old <- readMVal array
+    let oldList = vCast old
+    if null oldList
+        then return VUndef
+        else do
+            liftIO $ writeIORef (vCast array) $ VList $ fRem oldList
+            return $ fPick oldList
+
+op1Print f v = do
+    val <- readMVal v
+    vals <- mapM readMVal (vCast val)
+    let (handle, vs) = case vals of
+                        (VHandle h:vs)  -> (h, vs)
+                        _               -> (stdout, vals)
+    liftIO . f handle . concatMap vCast $ vs
+    return $ VBool True
+
+bool2n v = if v
+  then 1
+  else 0
+
+iboolIO f v = do
     ok <- liftIO $ (`catch` \_ -> return False) $ do
         f (vCast v)
         return True
+    return ok
+
+boolIO f v = do
+    ok <- iboolIO f v
     return $ VBool ok
 
 boolIO2 f u v = do
@@ -163,7 +219,7 @@ boolIO2 f u v = do
 opEval :: String -> Eval Val
 opEval str = do
     env <- ask
-    let env' = runRule env id ruleProgram str
+    let env' = runRule env id ruleProgram "<eval>" str
     val <- resetT $ local (\_ -> env') $ do
         evl <- asks envEval
         evl (envBody env')
@@ -177,6 +233,15 @@ mapStr f = map (chr . fromEnum . f)
 mapStr2 :: (Word8 -> Word8 -> Word8) -> [Word8] -> [Word8] -> String
 mapStr2 f x y = map (chr . fromEnum . uncurry f) $ x `zip` y
 
+mapStr2Fill :: (Word8 -> Word8 -> Word8) -> [Word8] -> [Word8] -> String
+mapStr2Fill f x y = map (chr . fromEnum . uncurry f) $ x `zipFill` y
+    where
+    zipFill [] [] = []
+    zipFill as [] = zip as (repeat 0)
+    zipFill [] bs = zip (repeat 0) bs
+    zipFill (a:as) (b:bs) = (a,b) : zipFill as bs
+
+
 op2 :: Ident -> Val -> Val -> Eval Val
 op2 "rename" = boolIO2 rename
 op2 "symlink" = boolIO2 createSymbolicLink
@@ -198,8 +263,8 @@ op2 "-"  = op2Numeric (-)
 op2 "~"  = op2Str (++)
 op2 "+|" = op2Int (.|.)
 op2 "+^" = op2Int xor
-op2 "~|" = op2Str $ mapStr2 (.|.)
-op2 "~^" = op2Str $ mapStr2 xor
+op2 "~|" = op2Str $ mapStr2Fill (.|.)
+op2 "~^" = op2Str $ mapStr2Fill xor
 op2 "=>" = \x y -> return $ VPair (x, y)
 op2 "cmp"= op2Ord vCastStr
 op2 "<=>"= op2Ord vCastRat
@@ -229,6 +294,8 @@ op2 "err"= op2 "//"
 op2 "nor"= op2 "!!"
 op2 "grep"= op2Grep
 op2 "map"= op2Map
+op2 "unshift" = op2Push (flip (++))
+op2 "push" = op2Push (++)
 op2 "split"= \x y -> return $ split (vCast x) (vCast y)
     where
     split :: VStr -> VStr -> Val
@@ -245,6 +312,14 @@ op2 "split"= \x y -> return $ split (vCast x) (vCast y)
 	| otherwise = (x:piece, rest') where (piece, rest') = breakOnGlue glue xs
 op2 s    = \x y -> return $ VError ("unimplemented binaryOp: " ++ s) (App s [] [Val x, Val y])
 
+op2Push f list _ = do
+    let (array:rest) = vCast list
+    old <- readMVal array
+    new <- mapM readMVal rest
+    let vals = vCast old `f` concatMap vCast new
+    liftIO $ writeIORef (vCast array) $ VList vals
+    return $ VInt $ genericLength vals
+
 op2Grep list sub@(VSub _) = op2Grep sub list
 op2Grep sub list = do
     vals <- (`filterM` vCast list) $ \x -> do
@@ -338,7 +413,11 @@ op2Numeric f x y
 primOp :: String -> String -> Params -> String -> Symbol
 primOp sym assoc prms ret = Symbol SOur name (Val sub)
     where
-    name = '&':'*':fixity ++ ':':sym
+    name | isAlpha (head sym)
+         , fixity == "prefix"
+         = "&*" ++ sym
+         | otherwise
+         = "&*" ++ fixity ++ (':':sym)
     sub  = VSub $ Sub { isMulti     = True
                       , subName     = sym
                       , subPad      = []
@@ -360,6 +439,7 @@ primOp sym assoc prms ret = Symbol SOur name (Val sub)
     symName = if modify then assoc ++ ":" ++ sym else sym
     (arity, fixity, modify) = case assoc of
         "pre"       -> (1, "prefix", False)
+        "spre"      -> (1, "prefix", False)
         "post"      -> (1, "postfix", True)
         "circum"    -> (1, "circumfix", True)
         "left"      -> (2, "infix", False)
@@ -369,11 +449,11 @@ primOp sym assoc prms ret = Symbol SOur name (Val sub)
         "list"      -> (0, "infix", False)
         other       -> (0, other, True)
         
-primDecl str = primOp sym assoc (reverse $ foldr foldParam [] prms) ret
+primDecl str = primOp sym assoc (foldr foldParam [] prms) ret
     where
     (ret:assoc:sym:prms') = words str
     takeWord = takeWhile isWord . dropWhile (not . isWord)
-    isWord = not . (`elem` "|(),")
+    isWord = not . (`elem` "|(),:")
     prms = map takeWord prms'
 
 doFoldParam cxt [] []       = [buildParam cxt "" "$?1" (Val VUndef)]
@@ -381,7 +461,8 @@ doFoldParam cxt [] (p:ps)   = (buildParam cxt "" (strInc $ paramName p) (Val VUn
 doFoldParam cxt (s:name) ps = (buildParam cxt [s] name (Val VUndef) : ps)
 
 foldParam :: String -> Params -> Params
-foldParam "List"    = doFoldParam "List" "*@?0"
+foldParam "List"    = doFoldParam "List" "*@?1"
+foldParam ('r':'w':'!':"List") = \ps -> ((buildParam "List" "" "@?0" (Val VUndef)) { isLValue = True }:ps)
 foldParam ('r':'w':'!':str) = \ps -> ((buildParam str "" "$?1" (Val VUndef)) { isLValue = True }:ps)
 foldParam ""        = id
 foldParam ('?':str) = \ps -> (buildParam "Num" "?" "$?1" (Val $ VNum (read def)):ps)
@@ -391,28 +472,39 @@ foldParam x         = doFoldParam x ""
 
 -- XXX -- Junctive Types -- XXX --
 
+-- spre is "symbolic pre", that is, operators for which a precedence has
+-- already been assigned in Parser.hs
+
+--    ret_val   assoc	op_name args
 initSyms = map primDecl . filter (not . null) . lines $ "\
-\\n   Bool      pre     !       (Bool)\
-\\n   Num       pre     +       (Num)\
-\\n   Num       pre     -       (Num)\
-\\n   Str       pre     ~       (Str)\
-\\n   Bool      pre     ?       (Bool)\
-\\n   List      pre     *       (List)\
-\\n   List      pre     **      (List)\
-\\n   Int       pre     +^      (Int)\
-\\n   Int       pre     ~^      (Str)\
-\\n   Bool      pre     ?^      (Bool)\
-\\n   Ref       pre     \\      (Any)\
-\\n   List      pre     ...     (Str|Num)\
+\\n   Bool      spre    !       (Bool)\
+\\n   Num       spre    +       (Num)\
+\\n   Num       spre    -       (Num)\
+\\n   Str       spre    ~       (Str)\
+\\n   Bool      spre    ?       (Bool)\
+\\n   Int       pre     int     (Int)\
+\\n   List      spre    *       (List)\
+\\n   List      spre    **      (List)\
+\\n   Int       spre    +^      (Int)\
+\\n   Int       spre    ~^      (Str)\
+\\n   Bool      spre    ?^      (Bool)\
+\\n   Ref       spre    \\      (Any)\
+\\n   List      spre    ...     (Str|Num)\
+\\n   Any       pre     undef   ()\
+\\n   Any       pre     undef   (rw!Any)\
 \\n   Any       post    ++      (rw!Num)\
 \\n   Num       post    --      (rw!Num)\
-\\n   Any       pre     ++      (rw!Num)\
-\\n   Num       pre     --      (rw!Num)\
+\\n   Any       spre    ++      (rw!Num)\
+\\n   Num       spre    --      (rw!Num)\
 \\n   Bool      pre     not     (Bool)\
 \\n   List      pre     map     (Code, List)\
 \\n   List      pre     grep    (Code, List)\
-\\n   List      pre     map     (List: Code)\
-\\n   List      pre     grep    (List: Code)\
+\\n   List      pre     map     (Array: Code)\
+\\n   List      pre     grep    (Array: Code)\
+\\n   Int       pre     push    (rw!Array, List)\
+\\n   Int       pre     unshift (rw!Array, List)\
+\\n   Scalar    pre     pop     (rw!Array)\
+\\n   Scalar    pre     shift   (rw!Array)\
 \\n   Str       pre     join    (List)\
 \\n   List      left    zip     (List)\
 \\n   List      pre     keys    (Hash)\
@@ -420,17 +512,20 @@ initSyms = map primDecl . filter (not . null) . lines $ "\
 \\n   List      pre     kv      (Hash)\
 \\n   Str       pre     perl    (List)\
 \\n   Any       pre     eval    (Str)\
+\\n   Any       pre     require (Str)\
 \\n   Any       pre     last    (?Num=1)\
 \\n   Any       pre     exit    (?Num=0)\
 \\n   Num       pre     rand    (?Num=1)\
+\\n   Bool      pre     defined (Any)\
+\\n   Str       pre     ref     (Any)\
 \\n   Num       pre     time    ()\
 \\n   Action    pre     print   (List)\
+\\n   Action    pre     say     (IO: List)\
 \\n   Action    pre     say     (List)\
 \\n   Action    pre     die     (List)\
 \\n   Any       pre     do      (Str)\
 \\n   IO        pre     open    (Str)\
 \\n   Any       pre     return  (Any)\
-\\n   Any       pre     <>      ()\
 \\n   Junction  pre     any     (List)\
 \\n   Junction  pre     all     (List)\
 \\n   Junction  pre     one     (List)\
@@ -447,10 +542,11 @@ initSyms = map primDecl . filter (not . null) . lines $ "\
 \\n   Bool      pre     rename  (Str, Str)\
 \\n   Bool      pre     symlink (Str, Str)\
 \\n   Bool      pre     link    (Str, Str)\
+\\n   Int       pre     unlink  (List)\
 \\n   Str       pre     readlink (Str)\
 \\n   List      pre     split   (Str, Str)\
-\\n   Str       pre     =       (IO)\
-\\n   List      pre     =       (IO)\
+\\n   Str       spre    =       (IO)\
+\\n   List      spre    =       (IO)\
 \\n   Junction  list    |       (List)\
 \\n   Junction  list    &       (List)\
 \\n   Junction  list    ^       (List)\
@@ -496,7 +592,7 @@ initSyms = map primDecl . filter (not . null) . lines $ "\
 \\n   Scalar    left    ^^      (Bool, Bool)\
 \\n   Scalar    left    //      (Bool, Bool)\
 \\n   List      list    ,       (List)\
-\\n   List      pre     <==     (List)\
+\\n   List      spre    <==     (List)\
 \\n   List      left    ==>     (List, Code)\
 \\n   Scalar    left    and     (Bool, Bool)\
 \\n   Scalar    left    or      (Bool, Bool)\
